module Main where

import Char
import System
import System.IO
import System.IO.Error
import System.IO.Unsafe
import System.Posix
import Data.IORef
import Foreign
import Monad
import Maybe
import Network
import Network.Socket (sIsConnected)
import Control.Concurrent
import Control.Exception hiding (catch, handle)
import Control.Concurrent.Chan

{-
children :: MVar [MVar ()]
children = unsafePerformIO (newMVar [])
    
waitForChildren :: IO ()
waitForChildren = do
  cs <- takeMVar children
  case cs of
       []   -> return ()
       m:ms -> do
         putMVar children ms
         takeMVar m
         waitForChildren
    
forkChild :: IO () -> IO ()
forkChild io = do
  mvar <- newEmptyMVar
  childs <- takeMVar children
  putMVar children (mvar:childs)
  forkIO (io `finally` putMVar mvar ())
  return ()
-}
maybeWhen :: (Monad m) => Bool -> m a -> m (Maybe a)
maybeWhen b m = if b then m >>= return . Just else return Nothing

data Event = Event { toThread   :: Maybe ThreadId,  -- Nothing means for everyone
                     fromThread :: ThreadId,
                     fromHost   :: String,
                     contents   :: EventContents }
           deriving (Eq, Show)

data EventContents = MsgEvent String
                   | QuitEvent
                   deriving (Eq, Show)

type EventHandler = Event -> IO ()

data ConnectionState = ConnectionState { handle :: Handle, 
                                         eventQueue :: Chan Event,
                                         host :: String }
--toString (MsgEvent from to m) = m

{-# NOINLINE eventStream #-}
{-# NOINLINE systemStream #-}
eventStream :: Chan Event
eventStream = unsafePerformIO newChan
systemStream :: Chan Event
systemStream = unsafePerformIO newChan

-- This is used to make sure that all client get the event, including
-- the client that broadcast.  Normally only used by the main thread.
broadcastEvent :: Event -> IO ()
broadcastEvent = writeChan eventStream

-- Clients should use this to communicate to the main thread via events
postEvent :: Chan Event -> Event -> IO ()
postEvent = writeChan 
postMsg :: String -> String -> IO ()
postMsg s host = do d <- readIORef dispatcherId
                    myid <- myThreadId
                    postEvent systemStream Event { fromThread = myid,
                                                   toThread = Just d,
                                                   fromHost = host,
                                                   contents = MsgEvent s }
postQuit :: String -> IO ()
postQuit host = do d <- readIORef dispatcherId
                   myid <- myThreadId
                   postEvent systemStream Event { fromThread = myid,
                                                  toThread = Just d,
                                                  fromHost = host,
                                                  contents = QuitEvent }
-- Clients can use this to grab their next event if they have one
nextEvent :: Chan Event -> IO (Maybe Event)
nextEvent c = do b <- moreEvents c
                 maybeWhen b $ readChan c

{-# NOINLINE dispatcherId #-}
dispatcherId :: IORef ThreadId
dispatcherId = unsafePerformIO $ do id <- myThreadId
                                    newIORef id
moreEvents :: Chan Event -> IO Bool
moreEvents c = fmap not $ isEmptyChan c

hGetNonBlocking :: Handle -> IO String
hGetNonBlocking h = allocaArray chunkSize (\p -> do 
                    n <- hGetBufNonBlocking h (p::Ptr Word8) chunkSize
                    s <- peekArray n p
                    rest <- if n == chunkSize then hGetNonBlocking h
                                              else return ""
                    return $ (map (chr.fromIntegral) s)++rest)
  where chunkSize = 255

-- TheHunter recommends: ReaderT (IORef s)

main = do installHandler sigPIPE Ignore Nothing
          args <- getArgs
          let port = if null args then error "No port given" 
                     else (read $ args!!0)::Int
          s <- listenOn $ PortNumber (fromIntegral port)
          forkIO $ do 
            tid <- myThreadId
            evaluate tid
            writeIORef dispatcherId tid
            mapM_ (\_ -> do (Event to from host c) <- readChan systemStream
                            let e = Event Nothing tid host c
                            if c == QuitEvent 
                              then broadcastEvent (Event (Just from) tid host c)
                              else broadcastEvent e) [1..]
          relisten s
     where
     relisten   s = mapM_ (\_ -> connection s) [1..]
     connection s = do (h, client, _) <- accept s
                       putStrLn $ "Accepted on: "++show h
                       forkIO $ do
                         eve <- dupChan eventStream
                         cs <- newIORef (ConnectionState { handle = h, 
                                                           host = client,
                                                           eventQueue = eve })
                         putStrLn $ "changing buffering on: "++show h
                         hSetBinaryMode h True 
                         hSetBuffering h LineBuffering
                         hSetBuffering stdout LineBuffering
                         ts <- fmap show myThreadId
                         putStrLn $ "Connection from "++client++" on: "++ts
                         catch (process cs) 
                               (\e -> do putStrLn $ "Caught: "++show e
                                         putStrLn $ "Closing socket."
                                         cleanUp cs)
       where
       process :: IORef ConnectionState -> IO ()
       process cs = mapM_ (\_ -> do input <- getInput cs
                                    processInput cs input
                                    handleEvent cs)
                          [1..]

getInput :: IORef ConnectionState -> IO String
getInput cs = handle `fmap` readIORef cs >>= hGetNonBlocking

processInput :: IORef ConnectionState -> String -> IO ()
processInput cs s = do host <- host `fmap` readIORef cs
                       when (not (null s) && (take 4 s) == "quit") (postQuit host)
                       unless (null s) (postMsg s host)

handleEvent :: IORef ConnectionState -> IO ()
handleEvent cs = do eq <- eventQueue `fmap` readIORef cs
                    e <- nextEvent eq
                    case e of 
                      Nothing -> return ()
                      (Just e') -> do tid <- myThreadId
                                      when (toThread e' == Nothing || 
                                            tid == fromJust (toThread e')) $ eventHandler cs e'

eventHandler :: IORef ConnectionState -> EventHandler
eventHandler cs (Event to from host c) = 
  case c of 
  (MsgEvent s) -> do h <- handle `fmap` readIORef cs
                     hPutStr h (host++": "++s)
  QuitEvent -> do h <- handle `fmap` readIORef cs
                  hPutStrLn h "QuitEvent"
                  cleanUp cs
                  exitWith ExitSuccess

cleanUp :: IORef ConnectionState -> IO ()
cleanUp cs = do h <- handle `fmap` readIORef cs
                hClose h
